package fos

import scala.util.parsing.input.Positional

/** Abstract Syntax Trees for terms. */
enum Term extends Positional:
  case True
  case False
  case Zero
  case Succ(t: Term)
  case Pred(t: Term)
  case IsZero(t: Term)
  case If(cond: Term, t1: Term, t2: Term)
  case Var(name: String)
  case Abs(v: String, tp: Type, t: Term)
  case App(t1: Term, t2: Term)
  case TermPair(t1: Term, t2: Term)
  case First(t: Term)
  case Second(t: Term)

  override def toString: String = this match
    case True => "true"
    case False => "false"
    case Zero => "0"
    case Succ(t: Term) => "succ " + t
    case Pred(t: Term) => "pred " + t
    case IsZero(t: Term) => "iszero " + t
    case If(cond: Term, t1: Term, t2: Term) => "if " + cond + " then " + t1 + " else " + t2
    case Var(name: String) => name
    case Abs(v: String, tp: Type, t: Term) => "(\\" + v + ":" + tp + "." + t + ")"
    case App(t1: Term, t2: Term) =>
      t1.toString + (t2 match
        case App(_, _) => " (" + t2.toString + ")" // left-associative
        case _         => " " + t2.toString
      )
    case TermPair(t1: Term, t2: Term) => "{" + t1 + "," + t2 + "}"
    case First(t: Term) => "fst " + t
    case Second(t: Term) => "snd " + t


/** Abstract Syntax Trees for types. */
enum Type extends Positional:

  case TypeBool
  case TypeNat
  case TypeFun(t1: Type, t2: Type)
  case TypePair(t1: Type, t2: Type)

  override def toString: String = this match 
    case TypeBool => "Bool"
    case TypeNat => "Nat"
    case TypeFun(t1: Type, t2: Type) =>
      (t1 match
        case TypeFun(_, _) => "(" + t1 + ")" // right-associative
        case _             => t1.toString
      ) + "->" + t2
    case TypePair(t1: Type, t2: Type) => "(" + t1 + " * " + t2 + ")"

